# written by Taylor Childers and Misha Salim
# usage: BALSAM_TEST_TOKENS=/path/to/file/with/tokens  BALSAM_TEST_SERVER=http://localhost:8000 locust
import copy
import logging
import os
import random
import time
from pathlib import Path
from typing import Any
from uuid import uuid4

import requests
import yaml
from locust import User, task  # type: ignore
from locust.env import Environment  # type: ignore
from locust.event import EventHook  # type: ignore

from balsam._api import models
from balsam.client import BasicAuthRequestsClient, OAuthRequestsClient

logger = logging.getLogger(__name__)

# get a token file
TEST_TOKENS_FILE = os.environ.get("BALSAM_TEST_TOKENS", "test-tokens.txt")
if not Path(TEST_TOKENS_FILE).is_file():
    raise RuntimeError(
        f"Cannot find token file {TEST_TOKENS_FILE}. Create this or set the path with BALSAM_TEST_TOKENS."
    )
TEST_BALSAM_SERVER = os.environ.get("BALSAM_TESET_SERVER", "http://0.0.0.0:8000")


# overload client using username/password token authentication
class LocustBalsamClientA(BasicAuthRequestsClient):
    def __init__(self, api_root: str, request_event: EventHook) -> None:
        # token file is typically here: `$HOME/.balsam/client.yml`
        with open(TEST_TOKENS_FILE) as fp:
            yml_file = yaml.load(fp, Loader=yaml.FullLoader)
            token = yml_file["token"]
        super().__init__(api_root, token=token)
        self._request_event = request_event

    def request(self, url: str, http_method: str, **kwargs: Any) -> Any:  # type: ignore
        start_time = time.perf_counter()
        url_parts = [(part if not part.isdigit() else ":id") for part in url.split("/")]
        url_name = "/".join(url_parts)
        request_meta = {
            "request_type": "balsam-requests-client",
            "name": f"{http_method} {url_name}",
            "response_length": 0,
            "response": None,
            "context": {},  # see HttpUser if you actually want to implement contexts
            "exception": None,
        }
        try:
            request_meta["response"] = super().request(url, http_method, **kwargs)
        except Exception as e:
            request_meta["exception"] = e
            raise
        finally:
            request_meta["response_time"] = (time.perf_counter() - start_time) * 1000
            self._request_event.fire(**request_meta)  # This is what makes the request actually get logged in Locust
        return request_meta["response"]


# using OAuth style authentication
class LocustBalsamClientB(OAuthRequestsClient):
    def __init__(self, api_root: str, request_event: EventHook) -> None:
        with open(TEST_TOKENS_FILE) as fp:
            tokens = [line.strip() for line in fp.readlines()]
        token = random.choice(tokens)
        tokens.remove(token)
        super().__init__(api_root, token=random.choice(tokens))
        self._request_event = request_event

    def request(self, url: str, http_method: str, **kwargs: Any) -> Any:  # type: ignore
        start_time = time.perf_counter()
        url_parts = [(part if not part.isdigit() else ":id") for part in url.split("/")]
        url_name = "/".join(url_parts)
        request_meta = {
            "request_type": "balsam-requests-client",
            "name": f"{http_method} {url_name}",
            "response_length": 0,
            "response": None,
            "context": {},  # see HttpUser if you actually want to implement contexts
            "exception": None,
        }
        try:
            request_meta["response"] = super().request(url, http_method, **kwargs)
        except Exception as e:
            request_meta["exception"] = e
            raise
        finally:
            request_meta["response_time"] = (time.perf_counter() - start_time) * 1000
            self._request_event.fire(**request_meta)  # This is what makes the request actually get logged in Locust
        return request_meta["response"]


# this is the locust User simiulator
class BalsamUser(User):  # type: ignore
    def __init__(self, environment: Environment) -> None:
        super().__init__(environment)
        # using username/password Authentication
        # change A to B to use OAuth instead
        self.client = LocustBalsamClientA(TEST_BALSAM_SERVER, environment.events.request)
        print("My token is ", self.client.token)
        self.Site = copy.deepcopy(models.Site)
        self.Site.objects = models.SiteManager(self.client)
        self.App = copy.deepcopy(models.App)  # Deepcopy
        self.App.objects = models.AppManager(self.client)
        self.BatchJob = copy.deepcopy(models.BatchJob)  # Deepcopy
        self.BatchJob.objects = models.BatchJobManager(self.client)
        self.Job = copy.deepcopy(models.Job)  # Deepcopy
        self.Job.objects = models.JobManager(self.client)
        self.TransferItem = copy.deepcopy(models.TransferItem)  # Deepcopy
        self.TransferItem.objects = models.TransferItemManager(self.client)
        self.Session = copy.deepcopy(models.Session)  # Deepcopy
        self.Session.objects = models.SessionManager(self.client)
        self.EventLog = copy.deepcopy(models.EventLog)  # Deepcopy
        self.EventLog.objects = models.EventLogManager(self.client)

    def on_start(self) -> None:
        name = f"test-site-{uuid4()}"
        path = Path(os.getcwd(), f"/projects/premade_site-{uuid4()}")
        self.premade_site = self.Site.objects.create(name=name, path=path)
        print("Created site id:", self.premade_site.id)
        assert self.premade_site.id is not None
        class_path = "premade_app"
        parameters = {"foo": {"required": False, "default": "foo"}, "bar": {"required": False, "default": "bar"}}
        self.premade_app = self.App.objects.create(
            site_id=self.premade_site.id, name=class_path, parameters=parameters, serialized_class="", source_code=""
        )
        print("Created app id:", self.premade_app.id, "at site:", self.premade_site.id)

    def on_stop(self) -> None:
        try:
            self.premade_app.delete()
        except requests.exceptions.HTTPError:
            logger.exception("failed to delete premade app")
        print("[", self.premade_site.id, "] sites: ", [site.id for site in self.Site.objects.all()])
        try:
            self.premade_site.delete()
        except requests.exceptions.HTTPError:
            logger.exception("failed to delete premade site")

    @task(3)
    def site_list(self) -> None:
        list(self.Site.objects.all())

    @task(1)
    def app_create(self) -> None:
        site = self.premade_site
        num_apps = len(self.App.objects.filter(site_id=site.id))
        parameters = {"foo": {"required": False, "default": "foo"}, "bar": {"required": False, "default": "bar"}}
        app = self.App.objects.create(
            site_id=site.id,
            name=f"Bar_{num_apps:08d}",
            serialized_class="",
            source_code="",
            parameters=parameters,
        )
        # delete app to cleanup
        time.sleep(1)
        app.delete()

    @task(3)
    def app_list(self) -> None:
        list(self.App.objects.all())

    @task(10)
    def job_create(self) -> None:
        app = self.premade_app

        parameters = {"foo": "yes", "bar": "maybe"}
        job_num = len(self.Job.objects.all())
        job = self.Job.objects.create("test/run_%08d" % job_num, app_id=app.id, parameters=parameters)

        # delete job to cleanup
        time.sleep(1)
        job.delete()

    @task(10)
    def job_list(self) -> None:
        list(self.Job.objects.all()[:5000])

    @task(10)
    def job_list_site(self) -> None:
        list(self.Job.objects.filter(site_id=self.premade_site.id))

    @task(10)
    def bulk_job_submission(self) -> None:
        site = self.premade_site
        app = self.premade_app
        # create a bunch of jobs:
        parameters = {"foo": "yes", "bar": "maybe"}
        batch_id = str(int(time.time()))
        # create some number of jobs
        num_jobs = random.choice(range(100, 500))
        jobs = [
            self.Job(
                f"batchjob_{batch_id}/run_{i:08d}",
                app_id=app.id,
                parameters=parameters,
                tags={"batch_id": batch_id},
            )
            for i in range(num_jobs)
        ]
        jobs = self.Job.objects.bulk_create(jobs)

        # move jobs to PREPROCESSED state
        for job in jobs:
            job.state = "PREPROCESSED"
        self.Job.objects.bulk_update(jobs)

        # create batch job
        simulated_nodes = 128
        batch_job = self.BatchJob.objects.create(
            site_id=site.id,
            project="datascience",
            queue="default",
            num_nodes=simulated_nodes,
            wall_time_min=30,
            job_mode="mpi",
        )
        # create session
        sess = self.Session.objects.create(batch_job_id=batch_job.id, site_id=site.id)

        # associate jobs with batchjob
        sess.acquire_jobs(
            max_num_jobs=1024,
            filter_tags={"batch_id": batch_id},
            states=["PREPROCESSED", "RESTART_READY"],
        )

        # session heartbeat
        sess.tick()

        # simulate runnings jobs in batches
        steps = int(len(jobs) / simulated_nodes) + 1
        for step in range(steps):
            # indices of jobs to operate on
            start = simulated_nodes * step
            end = min(simulated_nodes * (step + 1), len(jobs))

            # change jobs to running state
            for job in jobs[start:end]:
                job.state = "RUNNING"
            self.Job.objects.bulk_update(jobs[start:end])

            # session heartbeat
            sess.tick()

            # change jobs to done or error state
            for job in jobs[start:end]:
                job.state = "RUN_DONE" if random.random() < 0.95 else "RUN_ERROR"
            self.Job.objects.bulk_update(jobs[start:end])

            # session heartbeat
            sess.tick()

        # Delete batch job which also deletes jobs and session
        batch_job.delete()



import getpass
import json
import logging
from collections import Counter, defaultdict
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import dateutil.parser

from balsam.util import parse_to_utc

from .scheduler import (
    DelayedSubmitFail,
    SchedulerBackfillWindow,
    SchedulerJobLog,
    SchedulerJobStatus,
    SchedulerNonZeroReturnCode,
    SubprocessSchedulerInterface,
    scheduler_subproc,
)

PathLike = Union[Path, str]

logger = logging.getLogger(__name__)


def parse_cobalt_time_minutes(t_str: str) -> int:
    try:
        H, M, S = map(int, t_str.split(":"))
    except ValueError:
        return 0
    else:
        return H * 60 + M + round(S / 60)


class PBSScheduler(SubprocessSchedulerInterface):
    status_exe = "qstat"
    submit_exe = "qsub"
    delete_exe = "qdel"
    backfill_exe = "pbsnodes"

    # maps scheduler states to Balsam states
    _job_states = {
        "Q": "queued",
        "H": "queued",
        "T": "queued",
        "W": "queued",
        "S": "queued",
        "R": "running",
        "E": "running",
    }

    @staticmethod
    def _job_state_map(scheduler_state: str) -> str:
        return PBSScheduler._job_states.get(scheduler_state, "unknown")

    # maps Balsam status fields to the scheduler fields
    # should be a comprehensive list of scheduler status fields
    _status_fields = {
        "scheduler_id": "JobID",
        "state": "State",
        "wall_time_min": "WallTime",
        "queue": "Queue",
        "num_nodes": "Nodes",
        "project": "Project",
        "time_remaining_min": "TimeRemaining",
        "queued_time_min": "QueuedTime",
    }

    # when reading these fields from the scheduler apply
    # these maps to the string extracted from the output
    @staticmethod
    def _status_field_map(balsam_field: str) -> Optional[Callable[[str], Any]]:
        status_field_map: Dict[str, Callable[[str], Any]] = {
            "scheduler_id": lambda id: int(id),
            "state": PBSScheduler._job_state_map,
            "queue": lambda queue: str(queue),
            "num_nodes": lambda n: int(n),
            "time_remaining_min": parse_cobalt_time_minutes,
            "queued_time_min": parse_cobalt_time_minutes,
            "project": lambda project: str(project),
            "wall_time_min": parse_cobalt_time_minutes,
        }
        return status_field_map.get(balsam_field, None)

    # maps node list states to Balsam node states
    _node_states = {
        "busy": "busy",
        "idle": "idle",
        "cleanup-pending": "busy",
        "down": "busy",
        "allocated": "busy",
    }

    @staticmethod
    def _node_state_map(nodelist_state: str) -> str:
        try:
            return PBSScheduler._node_states[nodelist_state]
        except KeyError:
            logger.warning("node state %s is not recognized", nodelist_state)
            return "unknown"

    # maps the Balsam status fields to the node list fields
    # should be a comprehensive list of node list fields
    _nodelist_fields = {
        "id": "Node_id",
        "name": "Name",
        "queues": "Queues",
        "state": "Status",
        "mem": "MCDRAM",
        "numa": "NUMA",
        "wall_time_min": "Backfill",
    }

    # when reading these fields from the scheduler apply
    # these maps to the string extracted from the output
    @staticmethod
    def _nodelist_field_map(balsam_field: str) -> Callable[[str], Any]:
        nodelist_field_map = {
            "id": lambda id: int(id),
            "state": PBSScheduler._node_state_map,
            "queues": lambda x: x.split(":"),
            "wall_time_min": lambda x: parse_cobalt_time_minutes(x),
        }
        return nodelist_field_map.get(balsam_field, lambda x: x)

    @staticmethod
    def _render_submit_args(
        script_path: Union[Path, str], project: str, queue: str, num_nodes: int, wall_time_min: int, **kwargs: Any
    ) -> List[str]:
        hours = wall_time_min // 60
        minutes = wall_time_min - hours * 60
        args = [
            PBSScheduler.submit_exe,
            "-A",
            project,
            "-q",
            queue,
            "-l",
            f"select={num_nodes}",
            "-l",
            f"walltime={hours}:{minutes}:00",
            "-k",
            "doe",
            str(script_path),
        ]
        return args

    @staticmethod
    def _render_status_args(project: Optional[str], user: Optional[str], queue: Optional[str]) -> List[str]:
        pass

    @staticmethod
    def _render_delete_args(job_id: Union[int, str]) -> List[str]:
        return [PBSScheduler.delete_exe, str(job_id)]

    @staticmethod
    def _render_backfill_args() -> List[str]:
        return [PBSScheduler.backfill_exe, "-a", "-F", "json"]

    @staticmethod
    def _parse_submit_output(submit_output: str) -> int:
        try:
            return int(submit_output.split(".")[0])
        except Exception as exc:
            # Catch errors here and handle
            logger.warning(f"Exception: {exc}")
            raise

    @classmethod
    def get_statuses(
        cls,
        project: Optional[str] = None,
        user: Optional[str] = getpass.getuser(),
        queue: Optional[str] = None,
    ) -> Dict[int, SchedulerJobStatus]:
        # First call qstat to get user job ids
        args = [PBSScheduler.status_exe]
        stdout = scheduler_subproc(args)
        stdout_lines = [s for s in stdout.split("\n") if str(user) in s]
        if len(stdout_lines) == 0:
            return {}  # if there are no jobs in the queue return an empty dictionary
        user_job_ids = [s.split(".")[0] for s in stdout_lines]

        # Next call qstat to get job jsons
        args = [PBSScheduler.status_exe]
        args += user_job_ids
        args += "-f -F json".split()
        stdout = scheduler_subproc(args)
        stat_dict = cls._parse_status_output(stdout)
        return stat_dict

    @staticmethod
    def _parse_status_output(raw_output: str) -> Dict[int, SchedulerJobStatus]:
        # TODO: this can be much more efficient with a compiled regex findall()
        # logger.info(f"json status output {raw_output}")
        username = getpass.getuser()
        j = json.loads(raw_output)
        date_format = "%a %b %d %H:%M:%S %Y"
        status_dict = {}
        if "Jobs" in j.keys():
            try:
                for jobidstr, job in j["Jobs"].items():
                    # temporarily filter jobs by user due to PBS bug
                    job_username = job["Job_Owner"].split("@")[0]
                    if job_username != username:
                        continue
                    status = {}
                    try:
                        # array jobs can have a trailing "[]"; remove this
                        jobidstr = jobidstr.replace("[]", "")
                        jobid = int(jobidstr.split(".")[0])
                        status["scheduler_id"] = jobid
                    except ValueError:
                        logger.error(f"Error parsing jobid {jobidstr} in status output; skipping")
                        continue
                    status["state"] = PBSScheduler._job_states[job["job_state"]]  # type: ignore # noqa
                    status["time_remaining_min"] = 0
                    status["wall_time_min"] = 0
                    if "walltime" in job["Resource_List"].keys():
                        W = job["Resource_List"]["walltime"].split(":")
                        wall_time_min = int(W[0]) * 60 + int(W[1])  # 00:00:00
                        status["wall_time_min"] = wall_time_min
                        if status["state"] == "queued":  # type: ignore # noqa
                            status["time_remaining_min"] = wall_time_min
                        try:
                            if status["state"] == "running" and "stime" in job.keys():  # type: ignore # noqa
                                status["time_remaining_min"] = int(
                                    wall_time_min
                                    - (datetime.now() - datetime.strptime(job["stime"], date_format)).total_seconds()
                                    / 60
                                )
                        except Exception as err:
                            status["time_remaining_min"] = wall_time_min
                            logger.exception(f"Exception {str(err)} processing job {jobidstr} {job}")
                    status["queue"] = job["queue"]
                    status["num_nodes"] = job["Resource_List"]["nodect"]
                    status["project"] = job["project"]
                    status["queued_time_min"] = int(
                        (datetime.now() - datetime.strptime(job["qtime"], date_format)).total_seconds() / 60
                    )
                    status_dict[jobid] = SchedulerJobStatus(**status)
            except BaseException as err:
                logger.exception(f"Exception {str(err)} parsing {raw_output}")
        return status_dict

    @staticmethod
    def _parse_backfill_output(stdout: str) -> Dict[str, List[SchedulerBackfillWindow]]:
        # fill in this method later to support backfill
        # turam: input here will be json via "pbsnodes -a -F json"
        return dict()
        # prior cobalt impl follows
        raw_lines = stdout.strip().split("\n")
        nodelist = []
        node_lines = raw_lines[2:]
        logger.debug(node_lines)
        for line in raw_lines:
            try:
                line_dict = PBSScheduler._parse_nodelist_line(line)
            except (ValueError, TypeError):
                logger.debug(f"Cannot parse nodelist line: {line}")
            else:
                if line_dict["wall_time_min"] > 0 and line_dict["state"] == "idle":
                    nodelist.append(line_dict)

        windows = PBSScheduler._nodelist_to_backfill(nodelist)
        return windows

    @staticmethod
    def _parse_nodelist_line(line: str) -> Dict[str, Any]:
        fields = line.split()
        actual = len(fields)
        expected = len(PBSScheduler._nodelist_fields)

        if actual != expected:
            raise ValueError(f"Line has {actual} columns: expected {expected}:\n{fields}")

        status = {}
        for name, value in zip(PBSScheduler._nodelist_fields, fields):
            func = PBSScheduler._nodelist_field_map(name)
            status[name] = func(value)
        return status

    @staticmethod
    def _nodelist_to_backfill(
        nodelist: List[Dict[str, Any]],
    ) -> Dict[str, List[SchedulerBackfillWindow]]:
        queue_bf_times = defaultdict(list)
        windows = defaultdict(list)

        for entry in nodelist:
            bf_time = entry["wall_time_min"]
            queues = entry["queues"]
            for queue in queues:
                queue_bf_times[queue].append(bf_time)

        for queue, bf_times in queue_bf_times.items():
            # Mapping {bf_time: num_nodes}
            bf_counter = Counter(bf_times)
            # {
            #    queue_name: [(bf_time1, num_nodes1), (bf_time2, num_nodes2), ...],
            # }
            # For each queue, sorted with longer times first
            queue_bf_times[queue] = sorted(bf_counter.items(), reverse=True)

        for queue, bf_list in queue_bf_times.items():
            running_total = 0
            for bf_time, num_nodes in bf_list:
                running_total += num_nodes
                windows[queue].append(SchedulerBackfillWindow(num_nodes=running_total, wall_time_min=bf_time))
        return windows

    @staticmethod
    def _parse_time(line: str) -> datetime:
        time_str = line[: line.find("(UTC)")]
        return dateutil.parser.parse(time_str)

    @staticmethod
    def _parse_logs(scheduler_id: int, job_script_path: Optional[PathLike]) -> SchedulerJobLog:
        args = [PBSScheduler.status_exe]
        args += ["-x", "-f", "-F", "json"]
        args += [str(scheduler_id)]
        logger.info(f"_parse_logs issuing qstat: {str(args)}")
        try:
            stdout = scheduler_subproc(args)
        except SchedulerNonZeroReturnCode as e:
            if "Unknown Job Id" in str(e):
                logger.warning(f"Batch Job {scheduler_id} not found in PBS")
                raise DelayedSubmitFail
            return SchedulerJobLog()
        json_output = json.loads(stdout)
        # logger.info(f"_parse_logs json_output: {json_output}")
        if len(json_output["Jobs"]) == 0:
            logger.error("no job found for JOB ID = %s", scheduler_id)
            return SchedulerJobLog()
        job_data = list(json_output["Jobs"].values())[0]
        start_raw = job_data.get("stime")
        end_raw = job_data.get("mtime")
        if not (start_raw and end_raw):
            logger.warning(f"parse_logs got START_TIME: {start_raw}; FINISH_TIME: {end_raw}")
            return SchedulerJobLog()
        try:
            start = parse_to_utc(start_raw, local_zone="ET")
            end = parse_to_utc(end_raw, local_zone="ET")
        except dateutil.parser.ParserError:
            logger.warning(f"Failed to parse job_data times (START_TIME: {start_raw}) (FINISH_TIME: {end_raw})")
            return SchedulerJobLog()
        return SchedulerJobLog(start_time=start, end_time=end)

    @classmethod
    def discover_projects(cls) -> List[str]:
        """
        Get the user's allowed/preferred allocations
        Note: Could use sbank; currently uses Cobalt reporting of valid
              projects when an invalid project is given
        """
        """        click.echo("Checking with sbank for your current allocations...")
        with tempfile.NamedTemporaryFile() as fp:
            os.chmod(fp.name, 0o777)
            try:
                proc = subprocess.run(
                    "sbank projects -r polaris -f project_name --no-header --no-totals --no-sys-msg",
                    shell=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    encoding="utf-8",
                )
                print(f"proc is {proc}")
                sbank_out = proc.stdout
                projects = [p.strip() for p in sbank_out.split("\n") if p]
            except:
                projects = None

        """
        projects = None
        if not projects:
            projects = super().discover_projects()
        return projects


if __name__ == "__main__":
    pass
    """
    raw_output = open("qstat.out").read()
    status_dict = {}
    j = json.loads(raw_output)
    p = PBSScheduler()
    # scheduler_id = p.submit("hostname.sh","datascience","debug",1,10)
    # p.delete_job(scheduler_id)
    scheduler_id = p.submit("hostname.sh", "datascience", "debug", 1, 10)
    o = p.parse_logs(scheduler_id, "hostname")
    print("parse_logs:", o)
    o = p.get_statuses()
    for k, v in o.items():
        print(k, v)
    # not supporting this yet
    o = p.get_backfill_windows()
    print(o)
    pl = p.discover_projects()
    print(pl)
    """



import click
import yaml

from balsam.config import ClientSettings

from .utils import filter_by_sites


@click.group()
def app() -> None:
    """
    Manage Balsam applications
    """
    pass


@app.command()
@click.option("-v", "--verbose", is_flag=True)
@click.option("-s", "--site", "site_selector", default="")
def ls(site_selector: str, verbose: bool) -> None:
    """
    List Apps

    1) View apps across all sites

        balsam app ls --site=all

    2) Filter apps by specific site IDs or Path fragments

        balsam app ls --site=123,my_site_folder
    """
    client = ClientSettings.load_from_file().build_client()
    qs = client.App.objects.all()
    qs = filter_by_sites(qs, site_selector)
    if verbose:
        reprs = [yaml.dump(app.display_dict(), sort_keys=False, indent=4) for app in qs]
        print(*reprs, sep="\n----\n")
    else:
        sites = {site.id: site for site in client.Site.objects.all()}
        click.echo(f"{'ID':>5s}   {'Name':>18s}   {'Site':<20s}")
        apps = sorted(list(qs), key=lambda app: app.site_id)
        for a in apps:
            site = sites[a.site_id]
            click.echo(f"{a.id:>5d}   {a.name:>18s}   {site.name:<20s}")


@app.command()
@click.option("-n", "--name", "name_selector", default=None)
@click.option("-s", "--site", "site_selector", default="")
@click.option("-a", "--all", is_flag=True, default=False)
def rm(site_selector: str, name_selector: str, all: bool) -> None:
    """
    Remove Apps

    1) Remove named app

        balsam app rm -n hello_world

    1) Remove all apps across a site

        balsam app rm --all --site=123,my_site_folder

    2) Filter apps by specific site IDs or Path fragments

        balsam app rm -n hello_world --site=123,my_site_folder

    """
    client = ClientSettings.load_from_file().build_client()
    qs = client.App.objects.all()
    qs = filter_by_sites(qs, site_selector)

    if all and name_selector is not None:
        raise click.BadParameter("Specify app name or --all, but not both")
    elif not all and name_selector is None:
        raise click.BadParameter("Specify app name with -n or specify --all")
    else:
        app_list = []

        if all and site_selector == "":
            raise click.BadParameter("balsam app rm --all requires that you specify --site to remove jobs")
        elif all and site_selector != "":
            click.echo("THIS WILL DELETE ALL APPS IN SITE! CAUTION!")
            app_list = [a.name for a in list(qs)]
            num_apps = 0
            num_jobs = 0
        elif name_selector is not None:
            app_list = [name_selector]

        if len(app_list) > 0:
            for name in app_list:
                resolved_app = qs.get(name=name)
                resolved_id = resolved_app.id
                job_count = client.Job.objects.filter(app_id=resolved_id).count()

                if name_selector is not None:
                    appstr = f"App(id={resolved_id}, name={resolved_app.name}, site={resolved_app.site_id})"
                    if job_count == 0:
                        resolved_app.delete()
                        click.echo(f"Deleted {appstr}: there were no associated jobs.")
                    elif click.confirm(f"Really Delete {appstr}?? There are {job_count} Jobs that will be ERASED!"):
                        resolved_app.delete()
                        click.echo(f"Deleted App {resolved_id} ({name})")
                else:
                    num_apps += 1
                    num_jobs += job_count

            if all:
                if click.confirm(
                    f"Really DELETE {num_apps} apps and {num_jobs} jobs from site {site_selector}?? They will be ERASED!"
                ):
                    for name in app_list:
                        resolved_app = qs.get(name=name)
                        resolved_app.delete()
                    click.echo(f"Deleted {num_apps} apps and {num_jobs} jobs from site {site_selector}")
        else:
            click.echo("Found no apps to Delete")



import logging
import os

from balsam.platform.compute_node import PolarisNode

from .app_run import SubprocessAppRun

logger = logging.getLogger(__name__)


class PolarisRun(SubprocessAppRun):
    """
    https://www.open-mpi.org/doc/v3.0/man1/mpiexec.1.php
    """

    def _build_cmdline(self) -> str:
        node_ids = [h for h in self._node_spec.hostnames]

        # If the user does not set a cpu_bind option,
        # this code sets cpu-bind to be optimal for the gpus being used.
        # This does not handle the case where the application is using less than
        # 8 cpus per gpu.  This code will not skip the appropriate number of cpus
        # in the rank binding assignments.
        if "cpu_bind" in self._launch_params.keys():
            cpu_bind = self._launch_params.get("cpu_bind")
        elif "--cpu-bind" in self._launch_params.keys():
            cpu_bind = self._launch_params.get("--cpu-bind")
        else:
            # Here we grab the cpu_ids assigned to the job in the NodeSpec object
            # If this is not set in NodeSpec (it is only set for single node jobs),
            # then we take the cpu_id list from the Polaris ComputeNode subclass,
            # assuming the job will have use of all the cpus in nodes assigned to it.
            cpu_ids = self._node_spec.cpu_ids[0]
            polaris_node = PolarisNode(self._node_spec.node_ids[0], self._node_spec.hostnames[0])
            if not cpu_ids:
                cpu_ids = polaris_node.cpu_ids

            cpus_per_rank = self.get_cpus_per_rank()

            # PolarisNode reverses the order of the gpu_ids, so assigning the cpu-bind
            # in ascending cpu order is what we want here.
            cpu_bind_list = ["list"]
            for irank in range(self._ranks_per_node):
                cpu_bind_list.append(":")
                for i in range(cpus_per_rank):
                    if i > 0:
                        cpu_bind_list.append(",")
                    cid = str(cpu_ids[i + cpus_per_rank * irank])
                    cpu_bind_list.append(cid)
                    # If the job is using 2 hardware threads per core, we need to add those threads to the list
                    # The additional threads should go in the same ascending order (threads 0 and 32 are on the
                    # same physical core, threads 31 and 63 are on the same physical core)
                    if self._threads_per_core == 2:
                        cpu_bind_list.append(",")
                        cid = str(cpu_ids[i + cpus_per_rank * irank] + len(polaris_node.cpu_ids))
                        cpu_bind_list.append(cid)
            cpu_bind = "".join(cpu_bind_list)

        launch_params = []
        for k in self._launch_params.keys():
            if k != "cpu_bind" and k != "--cpu-bind":
                launch_params.append(str(self._launch_params[k]))

        # The value of -d depends on the setting of cpu_bind.  If cpu-bind=core, -d is the number of
        # physical cores per rank, otherwise it is the number of hardware threads per rank
        # https://docs.alcf.anl.gov/running-jobs/example-job-scripts/
        depth = self._threads_per_rank
        if "core" == cpu_bind:
            depth = self.get_cpus_per_rank()

        nid_str = ",".join(map(str, node_ids))
        args = [
            "mpiexec",
            "-np",
            self.get_num_ranks(),
            "-ppn",
            self._ranks_per_node,
            "--hosts",
            nid_str,
            "--cpu-bind",
            cpu_bind,
            "-d",
            depth,
            *launch_params,
            self._cmdline,
        ]
        return " ".join(str(arg) for arg in args)

    def _set_envs(self) -> None:
        envs = os.environ.copy()
        envs.update(self._envs)

        # Here we grab the gpus assigned to the job from NodeSpec.  NodeSpec only
        # sets this for single node jobs.  For multinode jobs, gpu_ids below will
        # be an empty list of lists (e.g. [[], []]).  The ordering of the gpu_ids
        # is reversed in PolarisNode and therefore the reverse ordering of
        # cpus to gpus should be reflected here
        gpu_ids = self._node_spec.gpu_ids[0]
        cpu_ids = self._node_spec.cpu_ids[0]
        logger.info(f"Polaris set_envs: gpu_ids={gpu_ids} cpu_ids={cpu_ids}")

        # Here we set CUDA_VISIBLE_DEVICES for single node jobs only.  We assume
        # for multinode jobs that the job has access to all gpus, and
        # CUDA_VISIBLE_DEVICES is set by the user, for example by local rank with an
        # gpu_affinity.sh script that wraps around the user application in the
        # ApplicationDefinition.
        # One special case: if your job has one node, 2 ranks, and 1 gpu per rank, the
        # code here will set CUDA_VISIBLE_DEVICES to "3,2" or "1,0".  A user provided
        # gpu_affinity.sh script should take this assigment and use it to reset
        # CUDA_VISIBLE_DEVICES for each local rank.  The user script should NOT
        # round-robin the setting CUDA_VISIBLE_DEVICES starting from 3.
        if gpu_ids:
            envs["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
            envs["CUDA_VISIBLE_DEVICES"] = ",".join(map(str, gpu_ids))
        envs["OMP_NUM_THREADS"] = str(self._threads_per_rank)
        self._envs = envs



from typing import Dict, List, cast

import click
import yaml

from balsam.schemas import BatchJobPartition, BatchJobState, JobMode

from .utils import filter_by_sites, load_client, load_site_from_selector, validate_partitions, validate_tags


@click.group()
def queue() -> None:
    """
    Submit and monitor BatchJobs (queued launcher pilots)
    """
    pass


@queue.command()
@click.option("-n", "--num-nodes", required=True, type=int)
@click.option("-t", "--wall-time-min", required=True, type=int)
@click.option("-q", "--queue", required=True)
@click.option("-A", "--project", required=True)
@click.option("-j", "--job-mode", type=click.Choice(["serial", "mpi"]), required=True)
@click.option("-tag", "--tag", "filter_tags", multiple=True, callback=validate_tags)
@click.option("-p", "--part", "partitions", multiple=True, callback=validate_partitions)
@click.option("-x", "--extra-param", "optional_params", multiple=True, callback=validate_tags)
@click.option("--site", "site_selector", default="")
def submit(
    num_nodes: int,
    wall_time_min: int,
    queue: str,
    project: str,
    job_mode: str,
    filter_tags: Dict[str, str],
    partitions: List[BatchJobPartition],
    optional_params: Dict[str, str],
    site_selector: str,
) -> None:
    """
    Submit a new BatchJob to the Site scheduler.

    1) Request 2 nodes to run all Jobs

        balsam queue submit -n 2 -t 60  -q default -A MyAllocation -j mpi

    2) Request 2 nodes to run only Jobs with certain tags

        balsam queue submit -n 2 -t 60  -q default -A alloc -j mpi -tag experiment=foo
    """
    client = load_client()
    BatchJob = client.BatchJob
    Site = client.Site

    try:
        site = load_site_from_selector(site_selector)
    except Site.DoesNotExist:
        raise click.BadParameter(f"No site matching --site {site_selector}")
    except Site.MultipleObjectsReturned:
        raise click.BadParameter(f"More than one site matching --site {site_selector}")
    assert site.id is not None

    job = BatchJob(
        site_id=site.id,
        num_nodes=num_nodes,
        wall_time_min=wall_time_min,
        queue=queue,
        project=project,
        job_mode=cast(JobMode, job_mode),
        optional_params=optional_params,
        filter_tags=filter_tags,
        partitions=partitions,
    )
    try:
        job.validate(
            site.allowed_queues,
            site.allowed_projects,
            site.optional_batch_job_params,
        )
    except ValueError as e:
        raise click.BadArgumentUsage(str(e))
    job.save()
    print(yaml.dump(job.display_dict(), sort_keys=False, indent=4))


@queue.command()
@click.option("-n", "--num", default=0, type=int)
@click.option("-h", "--history", is_flag=True, default=False)
@click.option("-v", "--verbose", is_flag=True, default=False)
@click.option("--site", "site_selector", default="")
@click.option("--scheduler_id", "scheduler_id", type=int, default=None)
def ls(history: bool, verbose: bool, num: int, site_selector: str, scheduler_id: int) -> None:
    """
    List BatchJobs

    1) View current BatchJobs

        balsam queue ls

    2) View historical BatchJobs at all sites

        balsam queue ls --history --site all

    3) View verbose record for BatchJob with scheduler id

        balsam queue ls --scheduler_id 12345 -v

    4) View the last n BatchJobs

        balsam queue ls --num n

    """
    client = load_client()
    BatchJob = client.BatchJob
    qs = filter_by_sites(BatchJob.objects.all(), site_selector)

    if not history and scheduler_id is None and num == 0:
        qs = qs.filter(state=["pending_submission", "queued", "running", "pending_deletion"])
        if len(qs) == 0:
            click.echo("No active batch jobs.  Use --history option to list completed batch jobs.")

    if scheduler_id is not None:
        qs = qs.filter(scheduler_id=scheduler_id)

    jobs = [j.display_dict() for j in qs]
    if not history and num > 0 and scheduler_id is None:
        click.echo(f"Displaying records for last {num} Batch Jobs")
        jobs = jobs[-num:]

    if verbose:
        for j in jobs:
            click.echo(j)
    else:
        sites = {site.id: site for site in client.Site.objects.all()}
        for job in jobs:
            site = sites[job["site_id"]]
            path_str = site.path.as_posix()
            if len(path_str) > 27:
                path_str = "..." + path_str[-27:]
            job["site"] = f"{site.name}"

        fields = [
            "id",
            "site",
            "scheduler_id",
            "state",
            "filter_tags",
            "project",
            "queue",
            "num_nodes",
            "wall_time_min",
            "job_mode",
        ]
        rows = [[str(j[field]) for field in fields] for j in jobs]

        col_widths = [len(f) for f in fields]
        for row in rows:
            for col_idx, width in enumerate(col_widths):
                col_widths[col_idx] = max(width, len(row[col_idx]))

        for i, field in enumerate(fields):
            fields[i] = field.rjust(col_widths[i] + 1)

        print(*fields)
        for row in rows:
            for i, col in enumerate(row):
                row[i] = col.rjust(col_widths[i] + 1)
            print(*row)


@queue.command()
@click.argument("scheduler_id", type=int)
def term(scheduler_id: int) -> None:
    """
    Terminate a BatchJob by scheduler id
    """
    client = load_client()
    BatchJob = client.BatchJob

    bjob = BatchJob.objects.get(scheduler_id=scheduler_id)
    if bjob.state in [BatchJobState.queued, BatchJobState.running]:
        bjob.state = BatchJobState.pending_deletion
        bjob.save()
        click.echo("Marked BatchJob for deletion")
    else:
        click.echo("BatchJob is not queued or running")



import os
from pathlib import Path
from typing import Optional, Set

PLATFORMS: Set[str] = {"alcf_theta", "alcf_thetagpu", "alcf_cooley", "alcf_polaris", "generic"}

LAUNCHER_STARTUP_TIMEOUT_SECONDS = {
    "generic": 30.0,
    "alcf_theta": 4 * 60.0 * 60,
    "alcf_thetagpu": 4 * 60.0 * 60,
    "alcf_cooley": 4 * 60.0 * 60,
    "alcf_polaris": 4 * 60.0 * 60,
}

LAUNCHER_SHUTDOWN_TIMEOUT_SECONDS = {
    "generic": 20.0,
    "alcf_theta": 120.0,
    "alcf_thetagpu": 120.0,
    "alcf_cooley": 120.0,
    "alcf_polaris": 120.0,
}


def get_platform() -> str:
    """
    Used to skip tests that are not applicable to the current platform
    """
    plat = os.environ.get("BALSAM_TEST_PLATFORM", "generic")
    return plat


def get_test_api_url() -> Optional[str]:
    """
    Test API server URL. If this is set, the Test DB is not used for
    API or integration tests (just run against a live server).
    """
    return os.environ.get("BALSAM_TEST_API_URL")


def get_test_db_url() -> str:
    """
    Test DB connection string. *Required* for standalone server tests.
    """
    return os.environ.get("BALSAM_TEST_DB_URL", "postgresql://postgres@localhost:5432/balsam-test")


def get_test_dir() -> Optional[str]:
    """
    Location for ephemeral Balsam test Sites.
    Must be mounted on launch, compute, and site-daemon nodes
    """
    return os.environ.get("BALSAM_TEST_DIR")


def get_test_log_dir() -> Path:
    """
    Log directory (persists as artifact after test session)
    Used by Live server (gunicorn).
    At the end of a Site integration test, the Site logs are copied
    into this directory as well.
    """
    base = os.environ.get("BALSAM_LOG_DIR") or Path.cwd()
    test_log_dir = Path(base).joinpath("pytest-logs")
    return test_log_dir


def get_launcher_startup_timeout() -> float:
    return LAUNCHER_STARTUP_TIMEOUT_SECONDS[get_platform()]


def get_launcher_shutdown_timeout() -> float:
    return LAUNCHER_SHUTDOWN_TIMEOUT_SECONDS[get_platform()]



import logging
import os
from datetime import timedelta
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseSettings, validator

from balsam.util import validate_log_level


class LoginMethod(str, Enum):
    oauth_authcode = "oauth_authcode"
    oauth_device = "oauth_device"
    password = "password"


class OAuthProviderSettings(BaseSettings):
    class Config:
        env_prefix = "balsam_oauth_"
        case_sensitive = False
        env_file = ".env"
        extra = "forbid"

    # Environment must export BALSAM_OAUTH_CLIENT_ID, etc...
    client_id: str
    client_secret: str
    redirect_scheme: str = "https"
    redirect_path: str = "/auth/oauth/ALCF/callback"
    request_uri: str = "https://oauth2-dev.alcf.anl.gov/o/authorize/"
    token_uri: str = "https://oauth2-dev.alcf.anl.gov/o/token/"
    user_info_uri: str = "https://oauth2-dev.alcf.anl.gov/api/v1/user/"
    scope: str = "read_basic_user_data"
    device_code_lifetime: timedelta = timedelta(seconds=300)
    device_poll_interval: timedelta = timedelta(seconds=3)


class AuthSettings(BaseSettings):
    class Config:
        env_prefix = "balsam_auth_"
        case_sensitive = False
        env_file = ".env"
        extra = "forbid"

    secret_key = (
        "3f4abcaa16a006de8a7ef520a4c77c7d51b3ddb235c1ee2b99187f0f049eaa76"  # WARNING: Use real secret in production
    )
    algorithm = "HS256"
    token_ttl: timedelta = timedelta(hours=48)
    auth_method: str = "user_from_token"
    login_methods: List[LoginMethod] = [LoginMethod.password]
    oauth_provider: Optional[OAuthProviderSettings] = None

    @validator("oauth_provider", always=True)
    def get_oauth_config(cls, v: Optional[OAuthProviderSettings]) -> Optional[OAuthProviderSettings]:
        if v is None and os.environ.get("BALSAM_OAUTH_CLIENT_ID"):
            return OAuthProviderSettings()
        return v


class Settings(BaseSettings):
    class Config:
        env_prefix = "balsam_"
        case_sensitive = False
        env_file = ".env"
        extra = "forbid"

    database_url: str = "postgresql://postgres@localhost:5432/balsam"
    auth: AuthSettings = AuthSettings()
    redis_params: Dict[str, Any] = {"unix_socket_path": "/tmp/redis-balsam.server.sock"}
    log_level: Union[str, int] = logging.INFO
    log_dir: Optional[Path]

    @validator("log_level", always=True)
    def validate_balsam_log_level(cls, v: Union[str, int]) -> int:
        return validate_log_level(v)

    @validator("log_dir", always=True)
    def validate_log_dir(cls, v: Optional[Path]) -> Optional[Path]:
        if v is None:
            return None
        if not v.exists():
            v.mkdir(parents=True, exist_ok=False)
        elif not v.is_dir():
            raise ValueError(f"{v} is not a valid log directory")
        v.chmod(0o700)
        return v

    def serialize_without_secrets(self) -> str:
        return self.json(
            exclude={
                "auth": {
                    "secret_key": ...,
                    "oauth_provider": {
                        "client_secret": ...,
                    },
                }
            }
        )



from typing import Any, Dict, List

from fastapi import APIRouter, Depends
from sqlalchemy import orm

from balsam import schemas
from balsam.server.auth import get_auth_method, get_webuser_session
from balsam.server.models import TransferItem, crud
from balsam.server.utils import Paginator

from .filters import TransferItemQuery

router = APIRouter()
auth = get_auth_method()


@router.get("/", response_model=schemas.PaginatedTransferItemOut)
def list(
    db: orm.Session = Depends(get_webuser_session),
    user: schemas.UserOut = Depends(auth),
    paginator: Paginator[TransferItem] = Depends(Paginator),
    q: TransferItemQuery = Depends(TransferItemQuery),
) -> Dict[str, Any]:
    """List data transfers associated with the user's Jobs."""
    count, transfers = crud.transfers.fetch(db, owner=user, paginator=paginator, filterset=q)
    return {"count": count, "results": transfers}


@router.get("/{transfer_id}", response_model=schemas.TransferItemOut)
def read(
    transfer_id: int, db: orm.Session = Depends(get_webuser_session), user: schemas.UserOut = Depends(auth)
) -> TransferItem:
    """Fetch a data transfer item by id."""
    count, transfers = crud.transfers.fetch(db, owner=user, transfer_id=transfer_id)
    assert isinstance(transfers, List)
    item: TransferItem = transfers[0]
    return item


@router.put("/{transfer_id}", response_model=schemas.TransferItemOut)
def update(
    transfer_id: int,
    data: schemas.TransferItemUpdate,
    db: orm.Session = Depends(get_webuser_session),
    user: schemas.UserOut = Depends(auth),
) -> schemas.TransferItemOut:
    """Update a transfer item by id."""
    updated_transfer = crud.transfers.update(db, owner=user, transfer_id=transfer_id, data=data)
    result = schemas.TransferItemOut.from_orm(updated_transfer)
    db.commit()
    return result


@router.patch("/", response_model=List[schemas.TransferItemOut])
def bulk_update(
    transfers: List[schemas.TransferItemBulkUpdate],
    db: orm.Session = Depends(get_webuser_session),
    user: schemas.UserOut = Depends(auth),
) -> List[schemas.TransferItemOut]:
    """Update a list of transfer items."""
    updated_transfers = crud.transfers.bulk_update(db, owner=user, update_list=transfers)
    result_transfers = [schemas.TransferItemOut.from_orm(t) for t in updated_transfers]
    db.commit()
    return result_transfers



import logging
import os
import subprocess
import time
from pathlib import Path, PosixPath
from typing import List, Optional, Sequence, Tuple, Union
from uuid import UUID

from globus_sdk import TransferData
from globus_sdk.exc import GlobusAPIError, GlobusConnectionError

from balsam.util.globus_auth import get_client

from .transfer import TaskInfo, TransferInterface, TransferRetryableError, TransferSubmitError

logger = logging.getLogger(__name__)


PathLike = Union[str, Path]
SrcDestRecursive = Tuple[PathLike, PathLike, bool]


def submit_sdk(src_endpoint: UUID, dest_endpoint: UUID, batch: Sequence[SrcDestRecursive]) -> str:
    client = get_client()
    transfer_data = TransferData(
        client,
        src_endpoint,
        dest_endpoint,
        label="",
        sync_level="size",
        verify_checksum=True,
        preserve_timestamp=False,
        encrypt_data=True,
        submission_id=None,
        delete_destination_extra=False,
        deadline=None,
        additional_fields=dict(
            notify_on_succeeded=False,
            notify_on_failed=False,
            notify_on_inactive=False,
        ),
    )
    for src, dest, recurse in batch:
        transfer_data.add_item(str(src), str(dest), recursive=recurse)
    try:
        res = client.submit_transfer(transfer_data)
    except GlobusAPIError as exc:
        raise TransferSubmitError(str(exc))
    task_id = res.get("task_id", None)
    if task_id is None:
        raise TransferSubmitError(str(res))
    logger.info(f"Submitted Globus transfer task {task_id}")
    return "globus:" + str(task_id)


def submit_subproc(src_endpoint: UUID, dest_endpoint: UUID, batch: List[SrcDestRecursive]) -> str:
    env = os.environ.copy()
    env["LC_ALL"] = "C.UTF-8"
    env["LANG"] = "C.UTF-8"
    batch_str = "\n".join(f"{src} {dest} {'-r' if recurse else ''}" for src, dest, recurse in batch)
    proc = subprocess.run(
        [
            "globus",
            "transfer",
            "--encrypt",
            "--batch",
            str(src_endpoint),
            str(dest_endpoint),
        ],
        input=batch_str,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding="utf-8",
        env=env,
        check=True,
    )
    for line in proc.stdout.split("\n"):
        if "Task ID" in line:
            return "globus:" + line.split()[-1]
    raise TransferSubmitError("Could not parse Task ID from submission")


class GlobusTransferInterface(TransferInterface):
    def __init__(self, endpoint_id: UUID, data_path: PathLike, endpoint_path: Optional[PathLike] = None):
        self.endpoint_id: UUID = UUID(str(endpoint_id))
        self.data_path: PathLike = data_path
        self.endpoint_path: Optional[PathLike] = endpoint_path

    @staticmethod
    def _state_map(status: str) -> str:
        return {
            "active": "active",
            "inactive": "inactive",
            "succeeded": "done",
            "failed": "error",
        }[status.strip().lower()]

    def submit_task(
        self,
        remote_loc: str,
        direction: str,
        transfer_paths: Sequence[Tuple[PathLike, PathLike, bool]],
    ) -> str:
        """Submit Transfer Task via Globus CLI"""
        transfer_paths_list: List[List[PathLike, PathLike, bool]] = []  # type: ignore [type-arg]
        if direction == "in":
            src_endpoint, dest_endpoint = UUID(str(remote_loc)), self.endpoint_id
            # modify destination path according to configured endpoint path
            if self.endpoint_path:
                for transfer in transfer_paths:
                    transfer_paths_list.append(
                        [
                            transfer[0],
                            PosixPath(str(transfer[1]).replace(str(self.data_path), str(self.endpoint_path))),
                            transfer[2],
                        ]
                    )
        elif direction == "out":
            src_endpoint, dest_endpoint = self.endpoint_id, UUID(str(remote_loc))
            # modify source path according to configured endpoint path
            if self.endpoint_path:
                for transfer in transfer_paths:
                    transfer_paths_list.append(
                        [
                            PosixPath(str(transfer[0]).replace(str(self.data_path), str(self.endpoint_path))),
                            transfer[1],
                            transfer[2],
                        ]
                    )
        else:
            raise ValueError("direction must be in or out")
        try:
            task_id = submit_sdk(src_endpoint, dest_endpoint, transfer_paths_list)  # type: ignore
        except TransferSubmitError as exc:
            if "ConsentRequired" in eval(exc.args[0]):
                logger.warn(
                    f"""Missing required data_access consent for Globus transfer.
 Ensure that you have given consent for Balsam to transfer with the required
 endpoints by executing the following command:
     balsam site globus-login -e {src_endpoint} -e {dest_endpoint}"""
                )
            raise
        except GlobusConnectionError as exc:
            raise TransferRetryableError(f"GlobusConnectionError in Transfer task submission: {exc}") from exc
        return str(task_id)

    @staticmethod
    def _poll_tasks(task_ids: Sequence[str]) -> List[TaskInfo]:
        client = get_client()
        try:
            task_list = [client.get_task(task_id) for task_id in task_ids]
            time.sleep(1)
        except GlobusConnectionError as exc:
            raise TransferRetryableError(f"GlobusConnectionError in client.get_task: {exc}")
        result = []
        for d in task_list:
            state = GlobusTransferInterface._state_map(d["status"])
            logger.debug(f"Mapping Globus task {d} to TransferItem state: {state}")
            info = {}
            if d.get("fatal_error"):
                info["error"] = d["fatal_error"]
            task_id = "globus:" + d["task_id"]
            result.append(TaskInfo(task_id=task_id, state=state, info=info))
        return result


TransferInterface._registry["globus"] = GlobusTransferInterface



from collections import defaultdict
from datetime import datetime
from itertools import chain
from logging import getLogger
from typing import Any, Dict, Iterable, List, Optional, Tuple

from fastapi import HTTPException, status
from sqlalchemy import Column, bindparam, func, insert, orm, select, update
from sqlalchemy.orm import Query, Session
from sqlalchemy.sql import Select

from balsam import schemas
from balsam.schemas.job import JobState, JobTransferItem
from balsam.server import ValidationError, models
from balsam.server.routers.filters import JobQuery
from balsam.server.utils import Paginator

logger = getLogger(__name__)


def owned_job_selector(owner: schemas.UserOut, columns: Optional[List[Column[Any]]] = None) -> Select:
    if columns is None:
        stmt = select(models.Job.__table__)  # type: ignore[arg-type]
    else:
        stmt = select(columns)
    return (  # type: ignore
        stmt.join(models.App.__table__, models.Job.app_id == models.App.id)  # type: ignore
        .join(models.Site.__table__, models.App.site_id == models.Site.id)
        .where(models.Site.owner_id == owner.id)
    )


def owned_job_query(db: Session, owner: schemas.UserOut, id_only: bool = False) -> "Query[models.Job]":
    qs: "Query[models.Job]" = db.query(models.Job.id) if id_only else db.query(models.Job)
    qs = qs.join(models.App).join(models.Site).filter(models.Site.owner_id == owner.id)  # type: ignore
    return qs


def populate_transfers(app: models.App, transfers: Dict[str, JobTransferItem]) -> List[Dict[str, Any]]:
    transfer_items = []
    for transfer_name, transfer_spec in transfers.items():
        transfer_slot = app.transfers.get(transfer_name)
        if transfer_slot is None:
            raise ValidationError(f"App {app.name} has no Transfer slot named {transfer_name}")
        direction, local_path = transfer_slot["direction"], transfer_slot["local_path"]
        recursive = transfer_slot["recursive"]
        assert direction in ["in", "out"]
        location_alias, remote_path = (
            transfer_spec.location_alias,
            transfer_spec.path.as_posix(),
        )
        url = app.site.transfer_locations.get(location_alias)
        if url is None:
            raise ValidationError(f"Site has no Transfer URL named {location_alias}")

        transfer_item = dict(
            direction=direction,
            local_path=local_path,
            remote_path=remote_path,
            recursive=recursive,
            location_alias=location_alias,
            state="pending" if direction == "in" else "awaiting_job",
            task_id="",
            transfer_info={},
        )
        transfer_items.append(transfer_item)
    return transfer_items


def fetch(
    db: Session,
    owner: schemas.UserOut,
    paginator: Optional[Paginator[models.Job]] = None,
    job_id: Optional[int] = None,
    filterset: Optional[JobQuery] = None,
) -> "Tuple[int, List[Dict[str, Any]]]":
    stmt = owned_job_selector(owner)
    if job_id is not None:
        stmt = stmt.where(models.Job.id == job_id)
    if filterset:
        stmt = filterset.apply_filters(stmt)
    if paginator is None:
        job = db.execute(stmt).mappings().one()
        return 1, [dict(job)]
    count_q = stmt.with_only_columns([func.count(models.Job.id)]).order_by(None)
    count = db.execute(count_q).scalar()
    stmt = paginator.paginate_core(stmt).order_by(models.Job.id)
    job_rows = [dict(j) for j in db.execute(stmt).mappings()]
    return count, job_rows


def bulk_create(
    db: Session, owner: schemas.UserOut, job_specs: List[schemas.ServerJobCreate]
) -> List[Dict[str, Any]]:
    now = datetime.utcnow()

    app_ids = set(job.app_id for job in job_specs)
    apps: Dict[int, models.App] = {
        app.id: app
        for app in db.query(models.App)
        .join(models.App.site)  # type: ignore
        .filter(models.Site.owner_id == owner.id, models.App.id.in_(app_ids))
        .options(orm.selectinload(models.App.site).load_only(models.Site.transfer_locations))
        .all()
    }
    if len(apps) < len(app_ids):
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Could not find one or more Apps. Double check app_id fields.",
        )

    parent_ids = set(pid for job in job_specs for pid in job.parent_ids)
    if parent_ids:
        parent_selector = owned_job_selector(owner, columns=[models.Job.id, models.Job.state])
        parent_selector = parent_selector.where(models.Job.id.in_(parent_ids))
        parent_states_by_id = {parent.id: parent.state for parent in db.execute(parent_selector)}
    else:
        parent_states_by_id = {}

    if len(parent_states_by_id) < len(parent_ids):
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Could not find one or more Jobs set as parents. Double check parent_ids fields.",
        )

    workdirs = set(job.workdir for job in job_specs)
    if len(workdirs) < len(job_specs):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Workdirs must be unique for each call to bulk-create Jobs.",
        )

    JobDict = Dict[str, Any]
    TransferList = List[Dict[str, Any]]
    jobs_by_workdir: Dict[str, JobDict] = {}
    transfer_lists_by_workdir: Dict[str, TransferList] = {}

    # Initial Job States depend on parents and any stage-in transfer items
    defaults = {
        "last_update": now,
        "pending_file_cleanup": True,
        "serialized_return_value": "",
        "serialized_exception": "",
        "batch_job_id": None,
    }
    for job_spec in job_specs:
        job_transfers = populate_transfers(apps[job_spec.app_id], job_spec.transfers)

        if any(parent_states_by_id[pid] != JobState.job_finished for pid in job_spec.parent_ids):
            state = JobState.awaiting_parents
        elif any(tr["direction"] == "in" for tr in job_transfers):
            state = JobState.ready
        else:
            state = JobState.staged_in

        # TODO: Removed jsonable_encoder for performance; what's needed to make the job_dict
        # SQLAlchemy safe?
        job_dict = job_spec.dict(exclude={"transfers", "workdir", "parent_ids"})
        workdir = job_spec.workdir.as_posix()
        job_dict.update(
            **defaults,
            state=state,
            workdir=workdir,
            parent_ids=list(job_spec.parent_ids),
        )

        jobs_by_workdir[workdir] = job_dict
        transfer_lists_by_workdir[workdir] = job_transfers

    # Bulk create jobs, RETURNING ids array and setting ids on job dicts
    stmt = insert(models.Job.__table__).returning(models.Job.workdir, models.Job.id)
    for res in db.execute(stmt, list(jobs_by_workdir.values())):
        jobs_by_workdir[res.workdir]["id"] = res.id

        for transfer_dict in transfer_lists_by_workdir[res.workdir]:
            transfer_dict["job_id"] = res.id

    # Bulk create transfer items
    transfers_flat_list = list(chain(*transfer_lists_by_workdir.values()))
    if transfers_flat_list:
        db.execute(insert(models.TransferItem.__table__), transfers_flat_list)

    # Bulk create events
    events = [
        dict(
            job_id=job["id"],
            timestamp=now,
            from_state=JobState.created,
            to_state=job["state"],
            data={},
        )
        for job in jobs_by_workdir.values()
    ]
    db.execute(insert(models.LogEvent.__table__), events)

    logger.debug(f"Bulk-created {len(jobs_by_workdir)} jobs")
    return list(jobs_by_workdir.values())


def _update_state(
    job: models.Job, state: str, state_timestamp: datetime, state_data: Dict[str, Any], transfer_items: List[Any]
) -> Tuple[Dict[str, Any], Dict[str, Any], List[int]]:
    if state == job.state or state is None:
        return {}, {}, []

    event = dict(
        job_id=job.id,
        from_state=job.state,
        to_state=state,
        timestamp=state_timestamp,
        data=state_data,
    )
    assert isinstance(event["data"], dict)

    update_dict: Dict[str, Any] = {"job_id": job.id, "state": state, "session_id": job.session_id}
    ready_transfers = []

    if state != "RUNNING" and job.session_id is not None:
        update_dict["session_id"] = None

    if state == "READY":
        if all(item.state == "done" for item in transfer_items if item.direction == "in"):
            update_dict["state"] = "STAGED_IN"
            event["data"]["message"] = "Skipped stage in"

    if state == "POSTPROCESSED":
        if all(item.state == "done" for item in transfer_items if item.direction == "out"):
            update_dict["state"] = "JOB_FINISHED"
            event["data"]["message"] = "Skipped stage out"
        else:
            for item in transfer_items:
                if item.state == "awaiting_job":
                    ready_transfers.append(item.id)

    if state == "STAGED_OUT":
        update_dict["state"] = "JOB_FINISHED"

    event["to_state"] = update_dict["state"]
    return update_dict, event, ready_transfers


def update_waiting_children(db: Session, finished_parent_ids: Iterable[int]) -> None:
    """
    When all of a job's parents reach JOB_FINISHED, update the job to READY
    """
    if not finished_parent_ids:
        return

    qs = (
        select((models.Job.__table__,))
        .where(models.Job.state == "AWAITING_PARENTS")
        .where(models.Job.parent_ids.overlap(finished_parent_ids))  # type: ignore[attr-defined]
    )
    children, transfer_items_by_jobid = select_jobs_for_update(db, qs, extra_cols=[models.Job.parent_ids])

    if not children:
        return

    parent_states = {pid: "JOB_FINISHED" for pid in finished_parent_ids}

    all_parent_ids = set(pid for job in children for pid in job.parent_ids)
    parent_ids_tofetch = all_parent_ids - set(finished_parent_ids)

    if parent_ids_tofetch:
        for job in db.execute(select((models.Job.id, models.Job.state)).where(models.Job.id.in_(parent_ids_tofetch))):
            parent_states[job.id] = job.state

    ready_children = []
    for child in children:
        if all(parent_states.get(pid) == "JOB_FINISHED" for pid in child.parent_ids):
            ready_children.append(child)

    now = datetime.utcnow()
    state_updates: List[Dict[str, Any]] = []
    events: List[Dict[str, Any]] = []

    for child in ready_children:
        state_update, event, _ = _update_state(
            job=child,
            state="READY",
            state_timestamp=now,
            state_data={"message": "All parents finished"},
            transfer_items=transfer_items_by_jobid[child.id],
        )
        if state_update:
            state_updates.append(state_update)
            events.append(event)

    if state_updates:
        db.execute(
            update(models.Job.__table__).where(models.Job.id == bindparam("job_id")),
            state_updates,
        )
        db.execute(insert(models.LogEvent.__table__), events)


def select_jobs_for_update(
    db: Session, qs: "Select", extra_cols: Optional[List[Column[Any]]] = None
) -> Tuple[List[Any], Dict[int, Any]]:
    """
    Select Jobs FOR UPDATE
    Return List[(job.id, job.state, job.session_id]) and Dict[job_id, (transfer.state, transfer.direction)]
    """
    if extra_cols is None:
        extra_cols = []
    qs = qs.with_only_columns([models.Job.id, models.Job.state, models.Job.session_id, *extra_cols])
    qs = qs.with_for_update(of=models.Job.__table__)
    jobs: List[Any] = db.execute(qs).all()
    job_ids = [job.id for job in jobs]

    transfer_items_by_job = defaultdict(list)
    for item in db.execute(
        select(
            (
                models.TransferItem.id,
                models.TransferItem.job_id,
                models.TransferItem.state,
                models.TransferItem.direction,
            )
        ).where(models.TransferItem.job_id.in_(job_ids))
    ):
        transfer_items_by_job[item.job_id].append(item)
    return jobs, transfer_items_by_job


def do_update_jobs(
    db: Session,
    update_jobs: List[Any],
    transfer_items_by_jobid: Dict[int, List[Any]],
    patch_dicts: Dict[int, Dict[str, Any]],
) -> None:
    events: List[Dict[str, Any]] = []
    ready_transfers: List[int] = []

    # First, perform job-wise updates:
    for job in update_jobs:
        update_data = patch_dicts[job.id]
        update_data["job_id"] = job.id
        if "workdir" in update_data:
            update_data["workdir"] = str(update_data["workdir"])

        state_update, event, update_transfer_ids = _update_state(
            job=job,
            state=update_data.pop("state", None),
            state_timestamp=update_data.pop("state_timestamp", datetime.utcnow()),
            state_data=update_data.pop("state_data", {}),
            transfer_items=transfer_items_by_jobid[job.id],
        )
        if state_update:
            update_data.update(state_update)
            events.append(event)
            ready_transfers.extend(update_transfer_ids)

    updates_list = list(patch_dicts.values())
    if updates_list:
        db.execute(
            update(models.Job.__table__).where(models.Job.id == bindparam("job_id")),
            updates_list,
        )
    if events:
        db.execute(insert(models.LogEvent.__table__), events)
    if ready_transfers:
        db.execute(
            update(models.TransferItem.__table__)
            .where(models.TransferItem.id.in_(ready_transfers))
            .values(state="pending")
        )

    # Then update affected children in a second query:
    finished_ids = [patch["job_id"] for patch in patch_dicts.values() if patch.get("state") == "JOB_FINISHED"]
    update_waiting_children(db, finished_ids)


def bulk_update(db: Session, owner: schemas.UserOut, patch_dicts: Dict[int, Dict[str, Any]]) -> int:
    job_ids = set(patch_dicts.keys())
    qs = owned_job_selector(owner).where(models.Job.id.in_(job_ids))
    update_jobs, transfer_items_by_jobid = select_jobs_for_update(db, qs)
    if len(update_jobs) < len(patch_dicts):
        raise ValidationError("Could not find some Job IDs")
    do_update_jobs(db, update_jobs, transfer_items_by_jobid, patch_dicts)
    return len(update_jobs)


def update_query(db: Session, owner: schemas.UserOut, update_data: Dict[str, Any], filterset: JobQuery) -> int:
    qs = owned_job_selector(owner)
    qs = filterset.apply_filters(qs)
    update_jobs, transfer_items_by_jobid = select_jobs_for_update(db, qs)
    if len(update_jobs) > schemas.MAX_ITEMS_PER_BULK_OP:
        raise HTTPException(
            status_code=400,
            detail=f"Cannot bulk-update more than {schemas.MAX_ITEMS_PER_BULK_OP} in a single API call.",
        )
    patch_dicts = {job.id: update_data.copy() for job in update_jobs}
    do_update_jobs(db, update_jobs, transfer_items_by_jobid, patch_dicts)
    return len(update_jobs)


def delete_query(
    db: Session,
    owner: schemas.UserOut,
    filterset: Optional[JobQuery] = None,
    job_id: Optional[int] = None,
) -> int:
    qs = owned_job_query(db, owner, id_only=True)
    if job_id is not None:
        qs = qs.filter(models.Job.id == job_id)
        qs.one()
    else:
        assert filterset is not None
        qs = filterset.apply_filters(qs)
    qs = qs.filter(models.Job.session_id.is_(None)).with_for_update(of=models.Job, skip_locked=True)  # type: ignore
    num_deleted = db.query(models.Job).filter(models.Job.id.in_(qs)).delete(synchronize_session=False)
    db.flush()
    logger.debug(f"Deleted {num_deleted} jobs")
    return num_deleted



import os

from .app_run import SubprocessAppRun


class SunspotRun(SubprocessAppRun):
    """
    https://www.open-mpi.org/doc/v3.0/man1/mpiexec.1.php
    """

    def _build_cmdline(self) -> str:
        node_ids = [h for h in self._node_spec.hostnames]
        cpu_bind = self._launch_params.get("cpu_bind", "none")
        nid_str = ",".join(map(str, node_ids))
        args = [
            "mpiexec",
            "-np",
            self.get_num_ranks(),
            "-ppn",
            self._ranks_per_node,
            "--hosts",
            nid_str,
            "--cpu-bind",
            cpu_bind,
            "-d",
            self._threads_per_rank,
            "--envall",
            self._cmdline,
        ]
        return " ".join(str(arg) for arg in args)

    # Overide default because sunspot does not use CUDA
    def _set_envs(self) -> None:
        envs = os.environ.copy()
        envs.update(self._envs)
        envs["OMP_NUM_THREADS"] = str(self._threads_per_rank)

        # Check the assigned GPU ID list from the first compute node:
        gpu_ids = self._node_spec.gpu_ids[0]
        if gpu_ids:
            envs["ZE_ENABLE_PCI_ID_DEVICE_ORDER"] = "1"
            envs["ZE_AFFINITY_MASK"] = ",".join(map(str, gpu_ids))

        self._envs = envs
